# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for creating and managing a polygon."""

from beartype import beartype as check_input_types
import numpy as np
from pint import Quantity
import pyvista as pv
from scipy.spatial.transform import Rotation as SpatialRotation

from ansys.geometry.core.math.matrix import Matrix33, Matrix44
from ansys.geometry.core.math.point import Point2D
from ansys.geometry.core.misc.measurements import DEFAULT_UNITS, Angle, Distance
from ansys.geometry.core.misc.units import UNITS
from ansys.geometry.core.sketch.face import SketchFace
from ansys.geometry.core.typing import Real


class Polygon(SketchFace):
    """Provides for modeling regular polygons.

    Parameters
    ----------
    center: Point2D
        Center point of the circle.
    inner_radius : ~pint.Quantity | Distance | Real
        Inner radius (apothem) of the polygon.
    sides : int
        Number of sides of the polygon.
    angle : ~pint.Quantity | Angle | Real, default: 0
        Placement angle for orientation alignment.
    """

    @check_input_types
    def __init__(
        self,
        center: Point2D,
        inner_radius: Quantity | Distance | Real,
        sides: int,
        angle: Quantity | Angle | Real = 0,
    ):
        """Initialize the polygon."""
        super().__init__()

        # Check the inputs
        self._center = center
        self._inner_radius = (
            inner_radius if isinstance(inner_radius, Distance) else Distance(inner_radius)
        )
        if self._inner_radius.value <= 0:
            raise ValueError("Radius must be a real positive value.")

        self._angle_offset = angle if isinstance(angle, Angle) else Angle(angle)

        # Verify that the number of sides is valid with preferred range
        if sides < 3:
            raise ValueError("The minimum number of sides to construct a polygon is three.")
        self._n_sides = sides

    @property
    def center(self) -> Point2D:
        """Center point of the polygon."""
        return self._center

    @property
    def inner_radius(self) -> Quantity:
        """Inner radius (apothem) of the polygon."""
        return self._inner_radius.value

    @property
    def n_sides(self) -> int:
        """Number of sides of the polygon."""
        return self._n_sides

    @property
    def angle(self) -> Quantity:
        """Orientation angle of the polygon."""
        return self._angle_offset.value

    @property
    def length(self) -> Quantity:
        """Side length of the polygon."""
        return 2 * self.inner_radius * np.tan(np.pi / self.n_sides)

    @property
    def outer_radius(self) -> Quantity:
        """Outer radius of the polygon."""
        return self.inner_radius / np.cos(np.pi / self.n_sides)

    @property
    def perimeter(self) -> Quantity:
        """Perimeter of the polygon."""
        return self.n_sides * self.length

    @property
    def area(self) -> Quantity:
        """Area of the polygon."""
        return (self.inner_radius * self.perimeter) / 2

    @property
    def visualization_polydata(self) -> pv.PolyData:
        """VTK polydata representation for PyVista visualization.

        The representation lies in the X/Y plane within
        the standard global Cartesian coordinate system.

        Returns
        -------
        pyvista.PolyData
            VTK pyvista.Polydata configuration.
        """
        # Compensate z orientation by -np.pi / 2 to match Geometry service polygon processing
        rotation = Matrix33(
            SpatialRotation.from_euler(
                "xyz",
                [0, 0, -np.pi / 2 + self.angle.m_as(UNITS.radian)],
                degrees=False,
            ).as_matrix()
        )

        transformation_matrix = Matrix44(
            [
                [
                    rotation[0, 0],
                    rotation[0, 1],
                    rotation[0, 2],
                    self.center.x.m_as(DEFAULT_UNITS.LENGTH),
                ],
                [
                    rotation[1, 0],
                    rotation[1, 1],
                    rotation[1, 2],
                    self.center.y.m_as(DEFAULT_UNITS.LENGTH),
                ],
                [
                    rotation[2, 0],
                    rotation[2, 1],
                    rotation[2, 2],
                    0,
                ],
                [0, 0, 0, 1],
            ]
        )

        return pv.Polygon(
            radius=self.inner_radius.m_as(DEFAULT_UNITS.LENGTH),
            n_sides=self.n_sides,
        ).transform(transformation_matrix)



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Trimmed curve class."""

from pint import Quantity

from ansys.api.geometry.v0.commands_pb2 import IntersectCurvesRequest
from ansys.api.geometry.v0.commands_pb2_grpc import CommandsStub
from ansys.geometry.core.connection.client import GrpcClient
from ansys.geometry.core.connection.conversions import trimmed_curve_to_grpc_trimmed_curve
from ansys.geometry.core.errors import protect_grpc
from ansys.geometry.core.math.point import Point3D
from ansys.geometry.core.shapes.curves.curve import Curve
from ansys.geometry.core.shapes.curves.curve_evaluation import CurveEvaluation
from ansys.geometry.core.shapes.parameterization import Interval
from ansys.geometry.core.typing import Real


class TrimmedCurve:
    """Represents a trimmed curve.

    A trimmed curve is a curve that has a boundary. This boundary comes in the form of an interval.

    Parameters
    ----------
    geometry : Curve
        Underlying mathematical representation of the curve.
    start : Point3D
        Start point of the curve.
    end : Point3D
        End point of the curve.
    interval : Interval
        Parametric interval that bounds the curve.
    length : Quantity
        Length of the curve.
    grpc_client : GrpcClient, default: None
        gRPC client that is required for advanced functions such as ``intersect_curve()``.
    """

    def __init__(
        self,
        geometry: Curve,
        start: Point3D,
        end: Point3D,
        interval: Interval,
        length: Quantity,
        grpc_client: GrpcClient = None,
    ):
        """Initialize ``TrimmedCurve`` class."""
        self._geometry = geometry
        self._start = start
        self._end = end
        self._interval = interval
        self._length = length
        self._grpc_client = grpc_client
        if grpc_client is not None:
            self._commands_stub = CommandsStub(self._grpc_client.channel)

    @property
    def geometry(self) -> Curve:
        """Underlying mathematical curve."""
        return self._geometry

    @property
    def start(self) -> Point3D:
        """Start point of the curve."""
        return self._start

    @property
    def end(self) -> Point3D:
        """End point of the curve."""
        return self._end

    @property
    @protect_grpc
    def length(self) -> Quantity:
        """Calculated length of the edge."""
        return self._length

    @property
    @protect_grpc
    def interval(self) -> Interval:
        """Interval of the curve that provides its boundary."""
        return self._interval

    def evaluate_proportion(self, param: Real) -> CurveEvaluation:
        """Evaluate the curve at a proportional value.

        A parameter of ``0`` corresponds to the start of the curve, while a parameter of ``1``
        corresponds to the end of the curve.

        Parameters
        ----------
        param : Real
            Parameter in the proportional range [0,1].

        Returns
        -------
        CurveEvaluation
            Resulting curve evaluation.
        """
        bounds = self.interval
        return self.geometry.evaluate(bounds.start + bounds.get_span() * param)

    def intersect_curve(self, other: "TrimmedCurve") -> list[Point3D]:
        """Get the intersect points of this trimmed curve with another one.

        If the two trimmed curves do not intersect, an empty list is returned.

        Parameters
        ----------
        other : TrimmedCurve
            Trimmed curve to intersect with.

        Returns
        -------
        list[Point3D]
            All points of intersection between the curves.
        """
        if self._grpc_client is None:
            raise ValueError(
                """Because this trimmed curve was not initialized with a gRPC client,
                the method cannot be called."""
            )
        first = trimmed_curve_to_grpc_trimmed_curve(self)
        second = trimmed_curve_to_grpc_trimmed_curve(other)
        res = self._commands_stub.IntersectCurves(
            IntersectCurvesRequest(first=first, second=second)
        )
        if res.intersect is False:
            return []
        return [Point3D([point.x, point.y, point.z]) for point in res.points]

    def __repr__(self) -> str:
        """Represent the trimmed curve as a string."""
        return (
            f"TrimmedCurve(geometry: {type(self.geometry)}, start: {self.start}, end: {self.end}, "
            f"interval: {self.interval}, length: {self.length})"
        )


class ReversedTrimmedCurve(TrimmedCurve):
    """Represents a reversed trimmed curve.

    When a curve is reversed, its start and end points are swapped, and parameters for evaluations
    are handled to provide expected results conforming to the direction of the curve. For example,
    evaluating a trimmed curve proportionally at 0 evaluates at the start point of the curve,
    but evaluating a reversed trimmed curve proportionally at 0 evaluates at what was previously
    the end point of the curve but is now the start point.

    Parameters
    ----------
    geometry : Curve
        Underlying mathematical representation of the curve.
    start : Point3D
        Original start point of the curve.
    end : Point3D
        Original end point of the curve.
    interval : Interval
        Parametric interval that bounds the curve.
    length : Quantity
        Length of the curve.
    grpc_client : GrpcClient, default: None
        gRPC client that is required for advanced functions such as `intersect_curve()`.
    """

    def __init__(
        self,
        geometry: Curve,
        start: Point3D,
        end: Point3D,
        interval: Interval,
        length: Quantity,
        grpc_client: GrpcClient = None,
    ):
        """Initialize an instance of a reversed trimmed curve."""
        super().__init__(geometry, start, end, interval, length, grpc_client)

        # Swap start and end points
        temp = self._start
        self._start = self._end
        self._end = temp

    def evaluate_proportion(self, param: Real) -> CurveEvaluation:  # noqa: D102
        # Evaluate starting from the end
        return self.geometry.evaluate(self.interval.end - self.interval.get_span() * param)



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for creating and managing a triangle."""

from beartype import beartype as check_input_types
import pyvista as pv

from ansys.geometry.core.math.point import Point2D
from ansys.geometry.core.misc.measurements import DEFAULT_UNITS
from ansys.geometry.core.sketch.face import SketchFace
from ansys.geometry.core.sketch.segment import SketchSegment


class Triangle(SketchFace):
    """Provides for modeling 2D triangles.

    Parameters
    ----------
    point1: Point2D
        Point that represents a triangle vertex.
    point2: Point2D
        Point that represents a triangle vertex.
    point3: Point2D
        Point that represents a triangle vertex.
    """

    @check_input_types
    def __init__(self, point1: Point2D, point2: Point2D, point3: Point2D):
        """Initialize the triangle."""
        super().__init__()

        self._point1 = point1
        self._point2 = point2
        self._point3 = point3

        # TODO: add plane to SketchSegment when available
        # https://github.com/ansys/pyansys-geometry/issues/1319
        self._edges.append(SketchSegment(self._point1, self._point2))
        self._edges.append(SketchSegment(self._point2, self._point3))
        self._edges.append(SketchSegment(self._point3, self._point1))

    @property
    def point1(self) -> Point2D:
        """Triangle vertex 1."""
        return self._point1

    @property
    def point2(self) -> Point2D:
        """Triangle vertex 2."""
        return self._point2

    @property
    def point3(self) -> Point2D:
        """Triangle vertex 3."""
        return self._point3

    @property
    def visualization_polydata(self) -> pv.PolyData:
        """VTK polydata representation for PyVista visualization.

        The representation lies in the X/Y plane within
        the standard global Cartesian coordinate system.

        Returns
        -------
        pyvista.PolyData
            VTK pyvista.Polydata configuration.
        """
        import numpy as np

        return pv.Triangle(
            np.array(
                [
                    [
                        self.point1.x.m_as(DEFAULT_UNITS.LENGTH),
                        self.point1.y.m_as(DEFAULT_UNITS.LENGTH),
                        0,
                    ],
                    [
                        self.point2.x.m_as(DEFAULT_UNITS.LENGTH),
                        self.point2.y.m_as(DEFAULT_UNITS.LENGTH),
                        0,
                    ],
                    [
                        self.point3.x.m_as(DEFAULT_UNITS.LENGTH),
                        self.point3.y.m_as(DEFAULT_UNITS.LENGTH),
                        0,
                    ],
                ],
                dtype=np.float64,
            )
        )



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides auxiliary math functions for PyAnsys Geometry."""

from beartype import beartype as check_input_types
import numpy as np

from ansys.geometry.core.typing import Real


@check_input_types
def get_two_circle_intersections(
    x0: Real, y0: Real, r0: Real, x1: Real, y1: Real, r1: Real
) -> tuple[tuple[Real, Real], tuple[Real, Real]] | None:
    """Get the intersection points of two circles.

    Parameters
    ----------
    x0 : Real
        x coordinate of the first circle.
    y0 : Real
        y coordinate of the first circle.
    r0 : Real
        Radius of the first circle.
    x1 : Real
        x coordinate of the second circle.
    y1 : Real
        y coordinate of the second circle.
    r1 : Real
        Radius of the second circle.

    Notes
    -----
    This function is based on the following StackOverflow post:
    https://stackoverflow.com/questions/55816902/finding-the-intersection-of-two-circles

    That post is based on the following implementation:
    https://paulbourke.net/geometry/circlesphere/

    Returns
    -------
    tuple[tuple[Real, Real], tuple[Real, Real]] | None
        Intersection points of the two circles if they intersect.
        The points are returned as ``((x3, y3), (x4, y4))``, where ``(x3, y3)`` and ``(x4, y4)``
        are the intersection points of the two circles. If the circles do not
        intersect, then ``None`` is returned.
    """
    # circle 1: (x0, y0), radius r0
    # circle 2: (x1, y1), radius r1
    d = np.sqrt((x1 - x0) ** 2 + (y1 - y0) ** 2)

    # non-intersecting
    if d > r0 + r1:
        return None
    # one circle within other
    if d < abs(r0 - r1):
        return None
    # coincident circles
    if d == 0 and r0 == r1:
        return None
    else:
        a = (r0**2 - r1**2 + d**2) / (2 * d)
        h = np.sqrt(r0**2 - a**2)
        x2 = x0 + a * (x1 - x0) / d
        y2 = y0 + a * (y1 - y0) / d

        dx = h * (y1 - y0) / d
        dy = h * (x1 - x0) / d
        x3 = x2 + dx
        y3 = y2 - dy

        x4 = x2 - dx
        y4 = y2 + dy

        return ((x3, y3), (x4, y4))



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides various option classes."""

from dataclasses import asdict, dataclass


@dataclass
class ImportOptions:
    """Import options when opening a file.

    Parameters
    ----------
    cleanup_bodies : bool = False
        Simplify geometry and clean up topology.
    import_coordinate_systems : bool = False
        Import coordinate systems.
    import_curves : bool = False
        Import curves.
    import_hidden_components_and_geometry : bool = False
        Import hidden components and geometry.
    import_names : bool = False
        Import names of bodies and curves.
    import_planes : bool = False
        Import planes.
    import_points : bool = False
        Import points.
    """

    cleanup_bodies: bool = False
    import_coordinate_systems: bool = False
    import_curves: bool = False
    import_hidden_components_and_geometry: bool = False
    import_names: bool = False
    import_planes: bool = False
    import_points: bool = False

    def to_dict(self):
        """Provide the dictionary representation of the ImportOptions class."""
        return {k: bool(v) for k, v in asdict(self).items()}



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""The problem area definition."""

from abc import abstractmethod
from typing import TYPE_CHECKING

from google.protobuf.wrappers_pb2 import Int32Value

from ansys.api.geometry.v0.repairtools_pb2 import (
    FixDuplicateFacesRequest,
    FixExtraEdgesRequest,
    FixInexactEdgesRequest,
    FixMissingFacesRequest,
    FixShortEdgesRequest,
    FixSmallFacesRequest,
    FixSplitEdgesRequest,
    FixStitchFacesRequest,
)
from ansys.api.geometry.v0.repairtools_pb2_grpc import RepairToolsStub
from ansys.geometry.core.connection import GrpcClient
from ansys.geometry.core.misc.auxiliary import (
    get_design_from_body,
    get_design_from_edge,
    get_design_from_face,
)
from ansys.geometry.core.misc.checks import check_type_all_elements_in_iterable
from ansys.geometry.core.tools.repair_tool_message import RepairToolMessage

if TYPE_CHECKING:  # pragma: no cover
    from ansys.geometry.core.designer.body import Body
    from ansys.geometry.core.designer.edge import Edge
    from ansys.geometry.core.designer.face import Face


class ProblemArea:
    """Represents problem areas.

    Parameters
    ----------
    id : str
        Server-defined ID for the problem area.
    grpc_client : GrpcClient
        Active supporting geometry service instance for design modeling.
    """

    def __init__(self, id: str, grpc_client: GrpcClient):
        """Initialize a new instance of a problem area class."""
        self._id = id
        self._id_grpc = Int32Value(value=int(id))
        self._repair_stub = RepairToolsStub(grpc_client.channel)

    @property
    def id(self) -> str:
        """The id of the problem area."""
        return self._id

    @abstractmethod
    def fix(self):
        """Fix problem area."""
        raise NotImplementedError("Fix method is not implemented in the base class.")


class DuplicateFaceProblemAreas(ProblemArea):
    """Provides duplicate face problem area definition.

    Represents a duplicate face problem area with unique identifier and associated faces.

    Parameters
    ----------
    id : str
        Server-defined ID for the body.
    grpc_client : GrpcClient
        Active supporting geometry service instance for design modeling.
    faces : list[Face]
        List of faces associated with the design.
    """

    def __init__(self, id: str, grpc_client: GrpcClient, faces: list["Face"]):
        """Initialize a new instance of the duplicate face problem areaclass."""
        super().__init__(id, grpc_client)

        from ansys.geometry.core.designer.face import Face

        # Verify that all elements in the list are of type Face
        check_type_all_elements_in_iterable(faces, Face)

        self._faces = faces

    @property
    def faces(self) -> list["Face"]:
        """The list of the edges connected to this problem area."""
        return self._faces

    def fix(self) -> RepairToolMessage:
        """Fix the problem area.

        Returns
        -------
        message: RepairToolMessage
            Message containing created and/or modified bodies.
        """
        if not self.faces:
            return RepairToolMessage(False, [], [])

        parent_design = get_design_from_face(self.faces[0])
        response = self._repair_stub.FixDuplicateFaces(
            FixDuplicateFacesRequest(duplicate_face_problem_area_id=self._id_grpc)
        )
        parent_design._update_design_inplace()
        message = RepairToolMessage(
            response.result.success,
            response.result.created_bodies_monikers,
            response.result.modified_bodies_monikers,
        )

        return message


class MissingFaceProblemAreas(ProblemArea):
    """Provides missing face problem area definition.

    Parameters
    ----------
    id : str
        Server-defined ID for the body.
    grpc_client : GrpcClient
        Active supporting geometry service instance for design modeling.
    edges : list[Edge]
        List of edges associated with the design.
    """

    def __init__(self, id: str, grpc_client: GrpcClient, edges: list["Edge"]):
        """Initialize a new instance of the missing face problem area class."""
        super().__init__(id, grpc_client)

        from ansys.geometry.core.designer.edge import Edge

        # Verify that all elements in the list are of type Edge
        check_type_all_elements_in_iterable(edges, Edge)

        self._edges = edges

    @property
    def edges(self) -> list["Edge"]:
        """The list of the edges connected to this problem area."""
        return self._edges

    def fix(self) -> RepairToolMessage:
        """Fix the problem area.

        Returns
        -------
        message: RepairToolMessage
            Message containing created and/or modified bodies.
        """
        if not self.edges:
            return RepairToolMessage(False, [], [])

        parent_design = get_design_from_edge(self.edges[0])
        response = self._repair_stub.FixMissingFaces(
            FixMissingFacesRequest(missing_face_problem_area_id=self._id_grpc)
        )
        parent_design._update_design_inplace()
        message = RepairToolMessage(
            response.result.success,
            response.result.created_bodies_monikers,
            response.result.modified_bodies_monikers,
        )
        return message


class InexactEdgeProblemAreas(ProblemArea):
    """Represents an inexact edge problem area with unique identifier and associated edges.

    Parameters
    ----------
    id : str
        Server-defined ID for the body.
    grpc_client : GrpcClient
        Active supporting geometry service instance for design modeling.
    edges : list[Edge]
        List of edges associated with the design.
    """

    def __init__(self, id: str, grpc_client: GrpcClient, edges: list["Edge"]):
        """Initialize a new instance of the inexact edge problem area class."""
        super().__init__(id, grpc_client)

        from ansys.geometry.core.designer.edge import Edge

        # Verify that all elements in the list are of type Edge
        check_type_all_elements_in_iterable(edges, Edge)

        self._edges = edges

    @property
    def edges(self) -> list["Edge"]:
        """The list of the edges connected to this problem area."""
        return self._edges

    def fix(self) -> RepairToolMessage:
        """Fix the problem area.

        Returns
        -------
        message: RepairToolMessage
            Message containing created and/or modified bodies.
        """
        if not self.edges:
            return RepairToolMessage(False, [], [])

        parent_design = get_design_from_edge(self.edges[0])
        response = self._repair_stub.FixInexactEdges(
            FixInexactEdgesRequest(inexact_edge_problem_area_id=self._id_grpc)
        )
        parent_design._update_design_inplace()
        message = RepairToolMessage(
            response.result.success,
            response.result.created_bodies_monikers,
            response.result.modified_bodies_monikers,
        )
        return message


class ExtraEdgeProblemAreas(ProblemArea):
    """Represents a extra edge problem area with unique identifier and associated edges.

    Parameters
    ----------
    id : str
        Server-defined ID for the body.
    grpc_client : GrpcClient
        Active supporting geometry service instance for design modeling.
    edges : list[Edge]
        List of edges associated with the design.
    """

    def __init__(self, id: str, grpc_client: GrpcClient, edges: list["Edge"]):
        """Initialize a new instance of the extra edge problem area class."""
        super().__init__(id, grpc_client)

        from ansys.geometry.core.designer.edge import Edge

        # Verify that all elements in the list are of type Edge
        check_type_all_elements_in_iterable(edges, Edge)

        self._edges = edges

    @property
    def edges(self) -> list["Edge"]:
        """The list of the ids of the edges connected to this problem area."""
        return self._edges

    def fix(self) -> RepairToolMessage:
        """Fix the problem area.

        Returns
        -------
        message: RepairToolMessage
            Message containing created and/or modified bodies.
        """
        if not self.edges:
            return RepairToolMessage(False, [], [])

        parent_design = get_design_from_edge(self.edges[0])
        request = FixExtraEdgesRequest(extra_edge_problem_area_id=self._id_grpc)
        response = self._repair_stub.FixExtraEdges(request)
        parent_design._update_design_inplace()
        message = RepairToolMessage(
            response.result.success,
            response.result.created_bodies_monikers,
            response.result.modified_bodies_monikers,
        )

        return message


class ShortEdgeProblemAreas(ProblemArea):
    """Represents a short edge problem area with a unique identifier and associated edges.

    Parameters
    ----------
    id : str
        Server-defined ID for the body.
    grpc_client : GrpcClient
        Active supporting geometry service instance for design modeling.
    edges : list[Edge]
        List of edges associated with the design.
    """

    def __init__(self, id: str, grpc_client: GrpcClient, edges: list["Edge"]):
        """Initialize a new instance of the ``ShortEdgeProblemAreas`` class."""
        super().__init__(id, grpc_client)

        from ansys.geometry.core.designer.edge import Edge

        # Verify that all elements in the list are edges
        check_type_all_elements_in_iterable(edges, Edge)

        self._edges = edges

    @property
    def edges(self) -> list["Edge"]:
        """The list of the ids of the edges connected to this problem area."""
        return self._edges

    def fix(self) -> RepairToolMessage:
        """Fix the problem area.

        Returns
        -------
        message: RepairToolMessage
            Message containing created and/or modified bodies.
        """
        if not self.edges:
            return RepairToolMessage(False, [], [])

        parent_design = get_design_from_edge(self.edges[0])
        response = self._repair_stub.FixShortEdges(
            FixShortEdgesRequest(short_edge_problem_area_id=self._id_grpc)
        )
        parent_design._update_design_inplace()
        message = RepairToolMessage(
            response.result.success,
            response.result.created_bodies_monikers,
            response.result.modified_bodies_monikers,
        )

        return message


class SmallFaceProblemAreas(ProblemArea):
    """Represents a small face problem area with a unique identifier and associated faces.

    Parameters
    ----------
    id : str
        Server-defined ID for the body.
    grpc_client : GrpcClient
        Active supporting geometry service instance for design modeling.
    faces : list[Face]
        List of edges associated with the design.
    """

    def __init__(self, id: str, grpc_client: GrpcClient, faces: list["Face"]):
        """Initialize a new instance of the small face problem area class."""
        super().__init__(id, grpc_client)

        from ansys.geometry.core.designer.face import Face

        # Verify that all elements in the list are of type Face
        check_type_all_elements_in_iterable(faces, Face)

        self._faces = faces

    @property
    def faces(self) -> list["Face"]:
        """The list of the ids of the edges connected to this problem area."""
        return self._faces

    def fix(self) -> RepairToolMessage:
        """Fix the problem area.

        Returns
        -------
        message: RepairToolMessage
            Message containing created and/or modified bodies.
        """
        if not self.faces:
            return RepairToolMessage(False, [], [])

        parent_design = get_design_from_face(self.faces[0])
        response = self._repair_stub.FixSmallFaces(
            FixSmallFacesRequest(small_face_problem_area_id=self._id_grpc)
        )
        parent_design._update_design_inplace()
        message = RepairToolMessage(
            response.result.success,
            response.result.created_bodies_monikers,
            response.result.modified_bodies_monikers,
        )
        return message


class SplitEdgeProblemAreas(ProblemArea):
    """Represents a split edge problem area with unique identifier and associated edges.

    Parameters
    ----------
    id : str
        Server-defined ID for the body.
    grpc_client : GrpcClient
        Active supporting geometry service instance for design modeling.
    edges : list[Edge]
        List of edges associated with the design.
    """

    def __init__(self, id: str, grpc_client: GrpcClient, edges: list["Edge"]):
        """Initialize a new instance of the split edge problem area class."""
        super().__init__(id, grpc_client)

        from ansys.geometry.core.designer.edge import Edge

        # Verify that all elements in the list are of type Edge
        check_type_all_elements_in_iterable(edges, Edge)

        self._edges = edges

    @property
    def edges(self) -> list["Edge"]:
        """The list of edges connected to this problem area."""
        return self._edges

    def fix(self) -> RepairToolMessage:
        """Fix the problem area.

        Returns
        -------
        message: RepairToolMessage
            Message containing created and/or modified bodies.
        """
        if not self.edges:
            return RepairToolMessage(False, [], [])

        parent_design = get_design_from_edge(self.edges[0])
        response = self._repair_stub.FixSplitEdges(
            FixSplitEdgesRequest(split_edge_problem_area_id=self._id_grpc)
        )
        parent_design._update_design_inplace()
        message = RepairToolMessage(
            response.result.success,
            response.result.created_bodies_monikers,
            response.result.modified_bodies_monikers,
        )
        return message


class StitchFaceProblemAreas(ProblemArea):
    """Represents a stitch face problem area with unique identifier and associated faces.

    Parameters
    ----------
    id : str
        Server-defined ID for the body.
    grpc_client : GrpcClient
        Active supporting geometry service instance for design modeling.
    bodies : list[Body]
        List of bodies associated with the design.
    """

    def __init__(self, id: str, grpc_client: GrpcClient, bodies: list["Body"]):
        """Initialize a new instance of the stitch face problem area class."""
        super().__init__(id, grpc_client)

        from ansys.geometry.core.designer.body import Body

        # Verify that all elements in the list are of type Body
        check_type_all_elements_in_iterable(bodies, Body)

        self._bodies = bodies

    @property
    def bodies(self) -> list["Body"]:
        """The list of the bodies connected to this problem area."""
        return self._bodies

    def fix(self) -> RepairToolMessage:
        """Fix the problem area.

        Returns
        -------
        message: RepairToolMessage
            Message containing created and/or modified bodies.
        """
        if not self.bodies:
            return RepairToolMessage(False, [], [])

        parent_design = get_design_from_body(self.bodies[0])
        response = self._repair_stub.FixStitchFaces(
            FixStitchFacesRequest(stitch_face_problem_area_id=self._id_grpc)
        )
        parent_design._update_design_inplace()
        message = RepairToolMessage(
            response.result.success,
            response.result.created_bodies_monikers,
            response.result.modified_bodies_monikers,
        )
        return message



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for creating and managing a slot."""

from beartype import beartype as check_input_types
import numpy as np
from pint import Quantity
import pyvista as pv
from scipy.spatial.transform import Rotation as SpatialRotation

from ansys.geometry.core.math.matrix import Matrix33
from ansys.geometry.core.math.point import Point2D
from ansys.geometry.core.misc.measurements import DEFAULT_UNITS, Angle, Distance
from ansys.geometry.core.misc.units import UNITS
from ansys.geometry.core.sketch.arc import Arc
from ansys.geometry.core.sketch.face import SketchFace
from ansys.geometry.core.sketch.segment import SketchSegment
from ansys.geometry.core.typing import Real


class Slot(SketchFace):
    """Provides for modeling a 2D slot.

    Parameters
    ----------
    center: :class:`Point2D <ansys.geometry.core.math.point.Point2D>`
        Center point of the slot.
    width : ~pint.Quantity | Distance | Real
        Width of the slot main body.
    height : ~pint.Quantity | Distance | Real
        Height of the slot.
    angle : ~pint.Quantity | Angle | Real, default: 0
        Placement angle for orientation alignment.
    """

    @check_input_types
    def __init__(
        self,
        center: Point2D,
        width: Quantity | Distance | Real,
        height: Quantity | Distance | Real,
        angle: Quantity | Angle | Real = 0,
    ):
        """Initialize the slot."""
        super().__init__()

        self._center = center
        self._width = width if isinstance(width, Distance) else Distance(width, center.unit)
        if self._width.value <= 0:
            raise ValueError("Width must be a real positive value.")
        width_magnitude = self._width.value.m_as(center.unit)

        self._height = height if isinstance(height, Distance) else Distance(height, center.unit)
        if self._height.value <= 0:
            raise ValueError("Height must be a real positive value.")
        height_magnitude = self._height.value.m_as(center.unit)

        if height_magnitude > width_magnitude:
            raise ValueError("Width must be greater than height.")

        if isinstance(angle, (int, float)):
            angle = Angle(angle, DEFAULT_UNITS.ANGLE)
        angle = angle if isinstance(angle, Angle) else Angle(angle, angle.units)

        rotation = Matrix33(
            SpatialRotation.from_euler(
                "xyz", [0, 0, angle.value.m_as(UNITS.radian)], degrees=False
            ).as_matrix()
        )

        half_h = height_magnitude / 2
        half_box_w = (width_magnitude - height_magnitude) / 2
        corner_1 = rotation @ [-half_box_w, half_h, 0]
        corner_2 = rotation @ [-half_box_w, -half_h, 0]
        corner_3 = rotation @ [half_box_w, -half_h, 0]
        corner_4 = rotation @ [half_box_w, half_h, 0]
        arc_1_center = rotation @ [-half_box_w, 0, 0]
        arc_2_center = rotation @ [half_box_w, 0, 0]

        self._slot_corner_1 = Point2D(
            [center.x.m + corner_1[0], center.y.m + corner_1[1]], center.unit
        )
        self._slot_corner_2 = Point2D(
            [center.x.m + corner_2[0], center.y.m + corner_2[1]], center.unit
        )
        self._slot_corner_3 = Point2D(
            [center.x.m + corner_3[0], center.y.m + corner_3[1]], center.unit
        )
        self._slot_corner_4 = Point2D(
            [center.x.m + corner_4[0], center.y.m + corner_4[1]], center.unit
        )
        self._arc_1_center = Point2D(
            [center.x.m + arc_1_center[0], center.y.m + arc_1_center[1]],
            center.unit,
        )
        self._arc_2_center = Point2D(
            [center.x.m + arc_2_center[0], center.y.m + arc_2_center[1]],
            center.unit,
        )

        # TODO: add plane to SketchSegment when available
        # https://github.com/ansys/pyansys-geometry/issues/1319
        self._arc1 = Arc(self._slot_corner_1, self._slot_corner_2, self._arc_1_center)
        self._segment1 = SketchSegment(self._slot_corner_2, self._slot_corner_3)
        self._arc2 = Arc(self._slot_corner_3, self._slot_corner_4, self._arc_2_center)
        self._segment2 = SketchSegment(self._slot_corner_4, self._slot_corner_1)

        self._edges.append(self._arc1)
        self._edges.append(self._segment1)
        self._edges.append(self._arc2)
        self._edges.append(self._segment2)

    @property
    def center(self) -> Point2D:
        """Center of the slot."""
        return self._center

    @property
    def width(self) -> Quantity:
        """Width of the slot."""
        return self._width.value

    @property
    def height(self) -> Quantity:
        """Height of the slot."""
        return self._height.value

    @property
    def perimeter(self) -> Quantity:
        """Perimeter of the slot."""
        return np.pi * self._height.value + 2 * (self._width.value - self._height.value)

    @property
    def area(self) -> Quantity:
        """Area of the slot."""
        return (
            np.pi * (self._height.value / 2) ** 2
            + (self._width.value - self._height.value) * self._height.value
        )

    @property
    def visualization_polydata(self) -> pv.PolyData:
        """VTK polydata representation for PyVista visualization.

        The representation lies in the X/Y plane within
        the standard global Cartesian coordinate system.

        Returns
        -------
        pyvista.PolyData
            VTK pyvista.Polydata configuration.
        """
        return pv.merge(
            [
                self._segment1.visualization_polydata,
                self._arc2.visualization_polydata,
                self._segment2.visualization_polydata,
                self._arc1.visualization_polydata,
            ]
        )



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module for connecting to instances of the Geometry service."""

import logging
import os
from pathlib import Path
from typing import TYPE_CHECKING

from ansys.geometry.core.connection.backend import ApiVersions, BackendType
from ansys.geometry.core.connection.client import MAX_MESSAGE_LENGTH
from ansys.geometry.core.connection.defaults import DEFAULT_PIM_CONFIG, DEFAULT_PORT
from ansys.geometry.core.connection.docker_instance import (
    _HAS_DOCKER,
    GeometryContainers,
    LocalDockerInstance,
)
from ansys.geometry.core.connection.product_instance import prepare_and_start_backend
from ansys.geometry.core.logger import LOG
from ansys.geometry.core.misc.checks import check_type, deprecated_argument

try:
    import ansys.platform.instancemanagement as pypim

    _HAS_PIM = True
except ModuleNotFoundError:  # pragma: no cover
    _HAS_PIM = False


if TYPE_CHECKING:  # pragma: no cover
    from ansys.geometry.core.modeler import Modeler


def launch_modeler(mode: str = None, **kwargs: dict | None) -> "Modeler":
    """Start the ``Modeler`` interface for PyAnsys Geometry.

    Parameters
    ----------
    mode : str, default: None
        Mode in which to launch the ``Modeler`` service. The default is ``None``,
        in which case the method tries to determine the mode automatically. The
        possible values are:

        * ``"pypim"``: Launches the ``Modeler`` service remotely using the PIM API.
        * ``"docker"``: Launches the ``Modeler`` service locally using Docker.
        * ``"geometry_service"``: Launches the ``Modeler`` service locally using the
          Ansys Geometry Service.
        * ``"spaceclaim"``: Launches the ``Modeler`` service locally using Ansys SpaceClaim.
        * ``"discovery"``: Launches the ``Modeler`` service locally using Ansys Discovery.

    **kwargs : dict, default: None
        Keyword arguments for the launching methods. For allowable keyword arguments, see the
        corresponding methods for each mode:

        * For ``"pypim"`` mode, see the :func:`launch_remote_modeler` method.
        * For ``"docker"`` mode, see the :func:`launch_docker_modeler` method.
        * For ``"geometry_service"`` mode, see the
          :func:`launch_modeler_with_geometry_service` method.
        * For ``"spaceclaim"`` mode, see the :func:`launch_modeler_with_spaceclaim` method.
        * For ``"discovery"`` mode, see the :func:`launch_modeler_with_discovery` method.

    Returns
    -------
    ansys.geometry.core.modeler.Modeler
        Pythonic interface for geometry modeling.

    Examples
    --------
    Launch the Geometry service.

    >>> from ansys.geometry.core import launch_modeler
    >>> modeler = launch_modeler()
    """
    if mode:
        return _launch_with_launchmode(mode, **kwargs)
    else:
        return _launch_with_automatic_detection(**kwargs)


def _launch_with_launchmode(mode: str, **kwargs: dict | None) -> "Modeler":
    """Start the ``Modeler`` interface for PyAnsys Geometry.

    Parameters
    ----------
    mode : str
        Mode in which to launch the ``Modeler`` service. The possible values are:

        * ``"pypim"``: Launches the ``Modeler`` service remotely using the PIM API.
        * ``"docker"``: Launches the ``Modeler`` service locally using Docker.
        * ``"geometry_service"``: Launches the ``Modeler`` service locally using the
          Ansys Geometry Service.
        * ``"spaceclaim"``: Launches the ``Modeler`` service locally using Ansys SpaceClaim.
        * ``"discovery"``: Launches the ``Modeler`` service locally using Ansys Discovery.

    **kwargs : dict, default: None
        Keyword arguments for the launching methods. For allowable keyword arguments, see the
        :func:`launch_remote_modeler` and :func:`launch_docker_modeler` methods. Some of these
        keywords might be unused.

    Returns
    -------
    ansys.geometry.core.modeler.Modeler
        Pythonic interface for geometry modeling.
    """
    # Ensure that the launch mode is a string
    if not isinstance(mode, str):
        raise TypeError("The launch mode must be a string.")

    # Ensure that the launch mode is lowercase
    mode = mode.lower()

    if mode == "pypim":
        return launch_remote_modeler(**kwargs)
    elif mode == "docker":
        return launch_docker_modeler(**kwargs)
    elif mode == "geometry_service":
        return launch_modeler_with_geometry_service(**kwargs)
    elif mode == "spaceclaim":
        return launch_modeler_with_spaceclaim(**kwargs)
    elif mode == "discovery":
        return launch_modeler_with_discovery(**kwargs)
    else:  # pragma: no cover
        raise ValueError(
            f"Invalid launch mode '{mode}'. The valid modes are: "
            "'pypim', 'docker', 'geometry_service', 'spaceclaim', 'discovery'."
        )


def _launch_with_automatic_detection(**kwargs: dict | None) -> "Modeler":
    """Start the ``Modeler`` interface based on automatic detection.

    Parameters
    ----------
    **kwargs : dict, default: None
        Keyword arguments for the launching methods. For allowable keyword arguments, see the
        :func:`launch_remote_modeler` and :func:`launch_docker_modeler` methods. Some of these
        keywords might be unused.

    Returns
    -------
    ansys.geometry.core.modeler.Modeler
        Pythonic interface for geometry modeling.
    """
    # This method is a wrapper for the other launching methods. It is used to
    # determine which method to call based on the environment.
    #
    # The order of the checks is as follows:
    #
    # 1. Check if PyPIM is configured and if the environment is configured for it.
    # 2. Check if Docker is installed and if the environment is configured for it.
    # 3. If you are on a Windows machine:
    #     - check if the Ansys Geometry service is installed.
    #     - check if Ansys SpaceClaim is installed.
    #     - check if Ansys Discovery is installed.

    # 1.Start PyAnsys Geometry with PyPIM if the environment is configured for it
    # and a directive on how to launch it was not passed.
    if _HAS_PIM and pypim.is_configured():
        LOG.info("Starting Geometry service remotely. The startup configuration is ignored.")
        return launch_remote_modeler(**kwargs)

    # Otherwise, we are in the "local Docker Container" scenario
    try:
        if _HAS_DOCKER and LocalDockerInstance.is_docker_installed():
            LOG.info("Starting Geometry service locally from Docker container.")
            return launch_docker_modeler(**kwargs)
    except Exception:
        LOG.warning(
            "The local Docker container could not be started."
            " Trying to start the Geometry service locally."
        )

    # If we are on a Windows machine, we can try to start the Geometry service locally,
    # through various methods: Geometry service, SpaceClaim, Discovery.
    if os.name == "nt":
        try:
            LOG.info("Starting Geometry service locally.")
            return launch_modeler_with_geometry_service(**kwargs)
        except Exception:
            LOG.warning(
                "The Geometry service could not be started locally."
                " Trying to start Ansys SpaceClaim locally."
            )

        try:
            LOG.info("Starting Ansys SpaceClaim with Geometry Service locally.")
            return launch_modeler_with_spaceclaim(**kwargs)
        except Exception:
            LOG.warning(
                "Ansys SpaceClaim could not be started locally."
                " Trying to start Ansys Discovery locally."
            )

        try:
            LOG.info("Starting Ansys Discovery with Geometry Service locally.")
            return launch_modeler_with_discovery(**kwargs)
        except Exception:
            LOG.warning("Ansys Discovery could not be started locally.")

    # If we reached this point...
    raise NotImplementedError("Geometry service cannot be initialized.")


def launch_remote_modeler(
    platform: str = "windows",
    version: str | None = None,
    client_log_level: int = logging.INFO,
    client_log_file: str | None = None,
    **kwargs: dict | None,
) -> "Modeler":
    """Start the Geometry service remotely using the PIM API.

    When calling this method, you must ensure that you are in an
    environment where `PyPIM <https://github.com/ansys/pypim>`_ is
    configured. You can use the
    :func:`pypim.is_configured <ansys.platform.instancemanagement.is_configured>`
    method to check if it is configured.

    Parameters
    ----------
    platform : str, default: None
        **Specific for Ansys Lab**. The platform option for the Geometry service.
        The default is ``"windows"``.
        This parameter is used to specify the operating system on which the
        Geometry service will run. The possible values are:

        * ``"windows"``: The Geometry service runs on a Windows machine.
        * ``"linux"``: The Geometry service runs on a Linux machine.

    version : str, default: None
        Version of the Geometry service to run in the three-digit format.
        For example, "232". If you do not specify the version, the server
        chooses the version.
    client_log_level : int, default: logging.INFO
        Log level for the client. The default is ``logging.INFO``.
    client_log_file : str, default: None
        Path to the log file for the client. The default is ``None``,
        in which case the client logs to the console.
    **kwargs : dict, default: None
        Placeholder to prevent errors when passing additional arguments that
        are not compatible with this method.

    Returns
    -------
    ansys.geometry.core.modeler.Modeler
        Instance of the Geometry service.
    """
    return _launch_pim_instance(
        is_pim_light=False,
        product_name="geometry",
        product_platform=platform,
        product_version=version,
        backend_type=None,
        client_log_level=client_log_level,
        client_log_file=client_log_file,
    )


def launch_docker_modeler(
    port: int = DEFAULT_PORT,
    connect_to_existing_service: bool = True,
    restart_if_existing_service: bool = False,
    name: str | None = None,
    image: GeometryContainers | None = None,
    client_log_level: int = logging.INFO,
    client_log_file: str | None = None,
    **kwargs: dict | None,
) -> "Modeler":
    """Start the Geometry service locally using Docker.

    When calling this method, a Geometry service (as a local Docker container)
    is started. By default, if a container with the Geometry service already exists
    at the given port, it connects to it. Otherwise, it tries to launch its own
    service.

    Parameters
    ----------
    port : int, optional
        Localhost port to deploy the Geometry service on or the
        the ``Modeler`` interface to connect to (if it is already deployed). By default,
        the value is the one for the ``DEFAULT_PORT`` connection parameter.
    connect_to_existing_service : bool, default: True
        Whether the ``Modeler`` interface should connect to a Geometry
        service already deployed at the specified port.
    restart_if_existing_service : bool, default: False
        Whether the Geometry service (which is already running)
        should be restarted when attempting connection.
    name : str, default: None
        Name of the Docker container to deploy. The default is ``None``,
        in which case Docker assigns it a random name.
    image : GeometryContainers, default: None
        The Geometry service Docker image to deploy. The default is ``None``,
        in which case the ``LocalDockerInstance`` class identifies the OS of your
        Docker engine and deploys the latest version of the Geometry service for
        that OS.
    client_log_level : int, default: logging.INFO
        Log level for the client. The default is ``logging.INFO``.
    client_log_file : str, default: None
        Path to the log file for the client. The default is ``None``,
        in which case the client logs to the console.
    **kwargs : dict, default: None
        Placeholder to prevent errors when passing additional arguments that
        are not compatible with this method.

    Returns
    -------
    Modeler
        Instance of the Geometry service.
    """
    from ansys.geometry.core.modeler import Modeler

    if not _HAS_DOCKER:  # pragma: no cover
        raise ModuleNotFoundError("The package 'docker' is required to use this function.")

    # Call the LocalDockerInstance ctor.
    docker_instance = LocalDockerInstance(
        port=port,
        connect_to_existing_service=connect_to_existing_service,
        restart_if_existing_service=restart_if_existing_service,
        name=name,
        image=image,
    )

    # Once the local Docker instance is ready... return the Modeler
    return Modeler(
        host="localhost",
        port=port,
        docker_instance=docker_instance,
        logging_level=client_log_level,
        logging_file=client_log_file,
    )


def launch_modeler_with_discovery_and_pimlight(
    version: str | None = None,
    client_log_level: int = logging.INFO,
    client_log_file: str | None = None,
    **kwargs: dict | None,
) -> "Modeler":
    """Start Ansys Discovery remotely using the PIM API.

    When calling this method, you must ensure that you are in an
    environment where `PyPIM <https://github.com/ansys/pypim>`_ is configured.
    You can use the :func:`pypim.is_configured <ansys.platform.instancemanagement.is_configured>`
    method to check if it is configured.

    Parameters
    ----------
    version : str, default: None
        Version of Discovery to run in the three-digit format.
        For example, "232". If you do not specify the version, the server
        chooses the version.
    client_log_level : int, default: logging.INFO
        Log level for the client. The default is ``logging.INFO``.
    client_log_file : str, default: None
        Path to the log file for the client. The default is ``None``,
        in which case the client logs to the console.
    **kwargs : dict, default: None
        Placeholder to prevent errors when passing additional arguments that
        are not compatible with this method.

    Returns
    -------
    ansys.geometry.core.modeler.Modeler
        Instance of Modeler.
    """
    return _launch_pim_instance(
        is_pim_light=True,
        product_name="discovery",
        product_version=version,
        backend_type=BackendType.DISCOVERY,
        client_log_level=client_log_level,
        client_log_file=client_log_file,
    )


def launch_modeler_with_geometry_service_and_pimlight(
    version: str | None = None,
    client_log_level: int = logging.INFO,
    client_log_file: str | None = None,
    **kwargs: dict | None,
) -> "Modeler":
    """Start the Geometry service remotely using the PIM API.

    When calling this method, you must ensure that you are in an
    environment where `PyPIM <https://github.com/ansys/pypim>`_ is configured.
    You can use the :func:`pypim.is_configured <ansys.platform.instancemanagement.is_configured>`
    method to check if it is configured.

    Parameters
    ----------
    version : str, default: None
        Version of the Geometry service to run in the three-digit format.
        For example, "232". If you do not specify the version, the server
        chooses the version.
    client_log_level : int, default: logging.INFO
        Log level for the client. The default is ``logging.INFO``.
    client_log_file : str, default: None
        Path to the log file for the client. The default is ``None``,
        in which case the client logs to the console.
    **kwargs : dict, default: None
        Placeholder to prevent errors when passing additional arguments that
        are not compatible with this method.

    Returns
    -------
    ansys.geometry.core.modeler.Modeler
        Instance of Modeler.
    """
    return _launch_pim_instance(
        is_pim_light=True,
        product_name="geometryservice",
        product_version=version,
        backend_type=BackendType.WINDOWS_SERVICE,
        client_log_level=client_log_level,
        client_log_file=client_log_file,
    )


def launch_modeler_with_spaceclaim_and_pimlight(
    version: str | None = None,
    client_log_level: int = logging.INFO,
    client_log_file: str | None = None,
    **kwargs: dict | None,
) -> "Modeler":
    """Start Ansys SpaceClaim remotely using the PIM API.

    When calling this method, you must ensure that you are in an
    environment where `PyPIM <https://github.com/ansys/pypim>`_ is configured.
    You can use the :func:`pypim.is_configured <ansys.platform.instancemanagement.is_configured>`
    method to check if it is configured.

    Parameters
    ----------
    version : str, default: None
        Version of SpaceClaim to run in the three-digit format.
        For example, "232". If you do not specify the version, the server
        chooses the version.
    client_log_level : int, default: logging.INFO
        Log level for the client. The default is ``logging.INFO``.
    client_log_file : str, default: None
        Path to the log file for the client. The default is ``None``,
        in which case the client logs to the console.
    **kwargs : dict, default: None
        Placeholder to prevent errors when passing additional arguments that
        are not compatible with this method.

    Returns
    -------
    ansys.geometry.core.modeler.Modeler
        Instance of Modeler.
    """
    return _launch_pim_instance(
        is_pim_light=True,
        product_name="scdm",
        product_version=version,
        backend_type=BackendType.SPACECLAIM,
        client_log_level=client_log_level,
        client_log_file=client_log_file,
    )


@deprecated_argument(arg="log_level", alternative="server_log_level")
@deprecated_argument(arg="logs_folder", alternative="server_logs_folder")
def launch_modeler_with_geometry_service(
    product_version: int = None,
    host: str = "localhost",
    port: int = None,
    enable_trace: bool = False,
    timeout: int = 60,
    server_log_level: int = 2,
    client_log_level: int = logging.INFO,
    server_logs_folder: str = None,
    client_log_file: str = None,
    log_level: int = None,  # DEPRECATED
    logs_folder: str = None,  # DEPRECATED
    **kwargs: dict | None,
) -> "Modeler":
    """Start the Geometry service locally using the ``ProductInstance`` class.

    When calling this method, a standalone Geometry service is started.
    By default, if an endpoint is specified (by defining `host` and `port` parameters)
    but the endpoint is not available, the startup will fail. Otherwise, it will try to
    launch its own service.

    Parameters
    ----------
    product_version: int, optional
        The product version to be started. Goes from v23.2.1 to
        the latest. Default is ``None``.
        If a specific product version is requested but not installed locally,
        a SystemError will be raised.

        **Ansys products versions and their corresponding int values:**

        * ``241`` : Ansys 24R1
    host: str, optional
        IP address at which the Geometry service will be deployed. By default,
        its value will be ``localhost``.
    port : int, optional
        Port at which the Geometry service will be deployed. By default, its
        value will be ``None``.
    enable_trace : bool, optional
        Boolean enabling the logs trace on the Geometry service console window.
        By default its value is ``False``.
    timeout : int, optional
        Timeout for starting the backend startup process. The default is 60.
    server_log_level : int, optional
        Backend's log level from 0 to 3:
            0: Chatterbox
            1: Debug
            2: Warning
            3: Error

        The default is ``2`` (Warning).
    client_log_level : int, optional
        Logging level to apply to the client. By default, INFO level is used.
        Use the logging module's levels: DEBUG, INFO, WARNING, ERROR, CRITICAL.
    server_logs_folder : str, optional
        Sets the backend's logs folder path. If nothing is defined,
        the backend will use its default path.
    client_log_file : str, optional
        Sets the client's log file path. If nothing is defined,
        the client will log to the console.
    log_level : int, optional
        DEPRECATED. Use ``server_log_level`` instead.
    logs_folder : str, optional
        DEPRECATED. Use ``server_logs_folder`` instead.
    **kwargs : dict, default: None
        Placeholder to prevent errors when passing additional arguments that
        are not compatible with this method.

    Raises
    ------
    ConnectionError
        If the specified endpoint is already in use, a connection
        error will be raised.
    SystemError
        If there is not an Ansys product 23.2 version or later installed
        a SystemError will be raised.

    Returns
    -------
    Modeler
        Instance of the Geometry service.

    Examples
    --------
    Starting a geometry service with the default parameters and getting back a ``Modeler``
    object:

    >>> from ansys.geometry.core import launch_modeler_with_geometry_service
    >>> modeler = launch_modeler_with_geometry_service()

    Starting a geometry service, on address ``10.171.22.44``, port ``5001``, with chatty
    logs, traces enabled and a ``300`` seconds timeout:

    >>> from ansys.geometry.core import launch_modeler_with_geometry_service
    >>> modeler = launch_modeler_with_geometry_service(host="10.171.22.44",
        port=5001,
        enable_trace= True,
        timeout=300,
        server_log_level=0)
    """
    # if api_version is passed, throw a warning saying that it is not used
    if "api_version" in kwargs:
        LOG.warning(
            "The 'api_version' parameter is not used in 'launch_modeler_with_geometry_service'. "
            "Please remove it from the arguments."
        )

    # If we are in a Windows environment, we are going to write down the server
    # logs in the %PUBLIC%/Documents/Ansys/GeometryService folder.
    if os.name == "nt" and server_logs_folder is None:
        # Writing to the "Public" folder by default - no write permissions specifically required.
        server_logs_folder = Path(os.getenv("PUBLIC"), "Documents", "Ansys", "GeometryService")
        LOG.info(f"Writing server logs to the default folder at {server_logs_folder}.")

    return prepare_and_start_backend(
        BackendType.WINDOWS_SERVICE,
        product_version=product_version,
        host=host,
        port=port,
        enable_trace=enable_trace,
        api_version=ApiVersions.LATEST,
        timeout=timeout,
        server_log_level=server_log_level,
        client_log_level=client_log_level,
        server_logs_folder=server_logs_folder,
        client_log_file=client_log_file,
        log_level=log_level,
        logs_folder=logs_folder,
    )


@deprecated_argument(arg="log_level", alternative="server_log_level")
def launch_modeler_with_discovery(
    product_version: int = None,
    host: str = "localhost",
    port: int = None,
    api_version: ApiVersions = ApiVersions.LATEST,
    timeout: int = 150,
    manifest_path: str = None,
    hidden: bool = False,
    server_log_level: int = 2,
    client_log_level: int = logging.INFO,
    client_log_file: str = None,
    log_level: int = None,  # DEPRECATED
    **kwargs: dict | None,
):
    """Start Ansys Discovery locally using the ``ProductInstance`` class.

    .. note::

       Support for Ansys Discovery is restricted to Ansys 24.1 onward.

    When calling this method, a standalone Discovery session is started.
    By default, if an endpoint is specified (by defining `host` and `port` parameters)
    but the endpoint is not available, the startup will fail. Otherwise, it will try to
    launch its own service.

    Parameters
    ----------
    product_version: int, optional
        The product version to be started. Goes from v23.2.1 to
        the latest. Default is ``None``.
        If a specific product version is requested but not installed locally,
        a SystemError will be raised.

        **Ansys products versions and their corresponding int values:**

        * ``241`` : Ansys 24R1
    host: str, optional
        IP address at which the Discovery session will be deployed. By default,
        its value will be ``localhost``.
    port : int, optional
        Port at which the Geometry service will be deployed. By default, its
        value will be ``None``.
    api_version: ApiVersions, optional
        The backend's API version to be used at runtime. Goes from API v21 to
        the latest. Default is ``ApiVersions.LATEST``.
    timeout : int, optional
        Timeout for starting the backend startup process. The default is 150.
    manifest_path : str, optional
        Used to specify a manifest file path for the ApiServerAddin. This way,
        it is possible to run an ApiServerAddin from a version an older product
        version.
    hidden : starts the product hiding its UI. Default is ``False``.
    server_log_level : int, optional
        Backend's log level from 0 to 3:
            0: Chatterbox
            1: Debug
            2: Warning
            3: Error

        The default is ``2`` (Warning).
    client_log_level : int, optional
        Logging level to apply to the client. By default, INFO level is used.
        Use the logging module's levels: DEBUG, INFO, WARNING, ERROR, CRITICAL.
    client_log_file : str, optional
        Sets the client's log file path. If nothing is defined,
        the client will log to the console.
    log_level : int, optional
        DEPRECATED. Use ``server_log_level`` instead.
    **kwargs : dict, default: None
        Placeholder to prevent errors when passing additional arguments that
        are not compatible with this method.

    Raises
    ------
    ConnectionError
        If the specified endpoint is already in use, a connection error will be raised.
    SystemError:
        If there is not an Ansys product 23.2 version or later installed
        or if a specific product's version is requested but not installed locally then
        a SystemError will be raised.

    Returns
    -------
    Modeler
        Instance of the Geometry service.

    Examples
    --------
    Starting an Ansys Discovery session with the default parameters and getting back a ``Modeler``
    object:

    >>> from ansys.geometry.core import launch_modeler_with_discovery
    >>> modeler = launch_modeler_with_discovery()

    Starting an Ansys Discovery V 23.2 session, on address ``10.171.22.44``, port ``5001``,
    with chatty logs, using API v231 and a ``300`` seconds timeout:

    >>> from ansys.geometry.core import launch_modeler_with_discovery
    >>> modeler = launch_modeler_with_discovery(product_version = 232,
        host="10.171.22.44",
        port=5001,
        api_version= 231,
        timeout=300,
        server_log_level=0)
    """
    for unused_var in ["server_logs_folder", "logs_folder"]:
        if unused_var in kwargs:
            LOG.warning(
                f"The '{unused_var}' parameter is not used in 'launch_modeler_with_discovery'. "
                "Please remove it from the arguments."
            )

    return prepare_and_start_backend(
        BackendType.DISCOVERY,
        product_version=product_version,
        host=host,
        port=port,
        enable_trace=False,
        api_version=api_version,
        timeout=timeout,
        manifest_path=manifest_path,
        hidden=hidden,
        server_log_level=server_log_level,
        client_log_level=client_log_level,
        client_log_file=client_log_file,
        log_level=log_level,
    )


@deprecated_argument(arg="log_level", alternative="server_log_level")
def launch_modeler_with_spaceclaim(
    product_version: int = None,
    host: str = "localhost",
    port: int = None,
    api_version: ApiVersions = ApiVersions.LATEST,
    timeout: int = 150,
    manifest_path: str = None,
    hidden: bool = False,
    server_log_level: int = 2,
    client_log_level: int = logging.INFO,
    client_log_file: str = None,
    log_level: int = None,  # DEPRECATED
    **kwargs: dict | None,
):
    """Start Ansys SpaceClaim locally using the ``ProductInstance`` class.

    When calling this method, a standalone SpaceClaim session is started.
    By default, if an endpoint is specified (by defining `host` and `port` parameters)
    but the endpoint is not available, the startup will fail. Otherwise, it will try to
    launch its own service.

    Parameters
    ----------
    product_version: int, optional
        The product version to be started. Goes from v23.2.1 to
        the latest. Default is ``None``.
        If a specific product version is requested but not installed locally,
        a SystemError will be raised.

        **Ansys products versions and their corresponding int values:**

        * ``232`` : Ansys 23R2 SP1
        * ``241`` : Ansys 24R1
    host: str, optional
        IP address at which the SpaceClaim session will be deployed. By default,
        its value will be ``localhost``.
    port : int, optional
        Port at which the Geometry service will be deployed. By default, its
        value will be ``None``.
    api_version: ApiVersions, optional
        The backend's API version to be used at runtime. Goes from API v21 to
        the latest. Default is ``ApiVersions.LATEST``.
    timeout : int, optional
        Timeout for starting the backend startup process. The default is 150.
    manifest_path : str, optional
        Used to specify a manifest file path for the ApiServerAddin. This way,
        it is possible to run an ApiServerAddin from a version an older product
        version.
    hidden : starts the product hiding its UI. Default is ``False``.
    server_log_level : int, optional
        Backend's log level from 0 to 3:
            0: Chatterbox
            1: Debug
            2: Warning
            3: Error

        The default is ``2`` (Warning).
    client_log_level : int, optional
        Logging level to apply to the client. By default, INFO level is used.
        Use the logging module's levels: DEBUG, INFO, WARNING, ERROR, CRITICAL.
    client_log_file : str, optional
        Sets the client's log file path. If nothing is defined,
        the client will log to the console.
    log_level : int, optional
        DEPRECATED. Use ``server_log_level`` instead.
    **kwargs : dict, default: None
        Placeholder to prevent errors when passing additional arguments that
        are not compatible with this method.

    Raises
    ------
    ConnectionError
        If the specified endpoint is already in use, a connection error will be raised.
    SystemError
        If there is not an Ansys product 23.2 version or later installed
        or if a specific product's version is requested but not installed locally then
        a SystemError will be raised.

    Returns
    -------
    Modeler
        Instance of the Geometry service.

    Examples
    --------
    Starting an Ansys SpaceClaim session with the default parameters and get back a ``Modeler``
    object:

    >>> from ansys.geometry.core import launch_modeler_with_spaceclaim
    >>> modeler = launch_modeler_with_spaceclaim()

    Starting an Ansys SpaceClaim V 23.2 session, on address ``10.171.22.44``, port ``5001``,
    with chatty logs, using API v231 and a ``300`` seconds timeout:

    >>> from ansys.geometry.core import launch_modeler_with_spaceclaim
    >>> modeler = launch_modeler_with_spaceclaim(product_version = 232,
        host="10.171.22.44",
        port=5001,
        api_version= 231,
        timeout=300,
        server_log_level=0)
    """
    for unused_var in ["server_logs_folder", "logs_folder"]:
        if unused_var in kwargs:
            LOG.warning(
                f"The '{unused_var}' parameter is not used in 'launch_modeler_with_spaceclaim'. "
                "Please remove it from the arguments."
            )

    return prepare_and_start_backend(
        BackendType.SPACECLAIM,
        product_version=product_version,
        host=host,
        port=port,
        enable_trace=False,
        api_version=api_version,
        timeout=timeout,
        manifest_path=manifest_path,
        hidden=hidden,
        server_log_level=server_log_level,
        client_log_level=client_log_level,
        client_log_file=client_log_file,
        log_level=log_level,
    )


def _launch_pim_instance(
    is_pim_light: bool,
    product_name: str,
    product_platform: str | None = None,
    product_version: str | None = None,
    backend_type: BackendType | None = None,
    client_log_level: int = logging.INFO,
    client_log_file: str | None = None,
):
    """
    Start `PyPIM <https://github.com/ansys/pypim>`_ using the PIM API.

    When calling this method, you must ensure that you are in an
    environment where PyPIM is configured. You can use the
    :func:`pypim.is_configured <ansys.platform.instancemanagement.is_configured>`
    method to check if it is configured.

    Parameters
    ----------
    is_pim_light : bool
        Whether PIM Light is being used. For example, whether PIM is
        running on a local machine.
    product_name : str
        Name of the service to run.
    product_platform : str, default: None
        Platform on which the service will run. **Specific for Ansys Lab**.
        This parameter is used to specify the operating system on which the
        Geometry service will run. The possible values are:

        * ``"windows"``: The Geometry service runs on a Windows machine.
        * ``"linux"``: The Geometry service runs on a Linux machine.

    product_version : str, default: None
        Version of the service to run.
    backend_type : BackendType, default: None
        Type of backend that PyAnsys Geometry is communicating with. By default, this
        value is unknown, which results in ``None`` being the default value.
    client_log_level : int, default: logging.INFO
        Log level for the client. The default is ``logging.INFO``.
    client_log_file : str, default: None
        Path to the log file for the client. The default is ``None``,
        in which case the client logs to the console.

    Returns
    -------
    ansys.geometry.core.modeler.Modeler
        Instance of the Geometry service.
    """
    from ansys.geometry.core.modeler import Modeler

    check_type(product_version, (type(None), str))

    if not _HAS_PIM:  # pragma: no cover
        raise ModuleNotFoundError(
            "The package 'ansys-platform-instancemanagement' is required to use this function."
        )

    # Platform is used mostly for Ansys Lab purposes. If product_version is defined, use it.
    # Higher priority is given to product_version.
    if product_platform:  # pragma: no cover
        if product_version:
            LOG.warning(
                "The 'product_platform' parameter is not used when 'product_version' is defined."
            )
        else:
            product_version = product_platform

    # If PIM Light is being used and PyPIM configuration is not defined... use defaults.
    if is_pim_light and not os.environ.get("ANSYS_PLATFORM_INSTANCEMANAGEMENT_CONFIG", None):
        os.environ["ANSYS_PLATFORM_INSTANCEMANAGEMENT_CONFIG"] = DEFAULT_PIM_CONFIG
        pop_out = True
    else:
        pop_out = False

    # Perform PyPIM connection
    pim = pypim.connect()
    instance = pim.create_instance(product_name=product_name, product_version=product_version)
    instance.wait_for_ready()
    channel = instance.build_grpc_channel(
        options=[
            ("grpc.max_receive_message_length", MAX_MESSAGE_LENGTH),
        ]
    )

    # If the default PyPIM configuration was used... remove
    if pop_out:
        os.environ.pop("ANSYS_PLATFORM_INSTANCEMANAGEMENT_CONFIG")

    return Modeler(
        channel=channel,
        remote_instance=instance,
        backend_type=backend_type,
        logging_level=client_log_level,
        logging_file=client_log_file,
    )
